@@ -13,7 +13,7 @@
  *
  * You should have received a copy of the GNU Lesser General
  * Public License along with this library; if not, see
- * <http://www.gnu.org/licenses/>.
+ * <https://www.gnu.org/licenses/>.
  */
 
 #include "config.h"
@@ -24,8 +24,11 @@
 #define NEEDS_BABL_DB
 #include "babl-internal.h"
 #include "babl-db.h"
+#include "babl-ref-pixels.h"
 
-static int babl_format_destruct (void *babl)
+
+static int 
+babl_format_destruct (void *babl)
 {
   BablFormat *format = babl;
   if (format->image_template != NULL)
@@ -41,14 +44,16 @@ static int babl_format_destruct (void *babl)
 }
 
 static Babl *
-format_new (const char     *name,
-            int             id,
-            int             planar,
-            int             components,
-            BablModel      *model,
-            BablComponent **component,
-            BablSampling  **sampling,
-            const BablType **type)
+format_new (const char      *name,
+            int              id,
+            int              planar,
+            int              components,
+            BablModel       *model,
+            const Babl      *space,
+            BablComponent  **component,
+            BablSampling   **sampling,
+            const BablType **type,
+            const char      *doc)
 {
   Babl *babl;
 
@@ -82,6 +87,7 @@ format_new (const char     *name,
                       sizeof (BablType *) * (components) +
                       sizeof (int) * (components) +
                       sizeof (int) * (components));
+
   babl_set_destructor (babl, babl_format_destruct);
 
   babl->format.from_list = NULL;
@@ -95,9 +101,13 @@ format_new (const char     *name,
 
   strcpy (babl->instance.name, name);
 
-  babl->format.model      = model;
   babl->format.components = components;
 
+  if (space == babl_space ("sRGB"))
+    babl->format.model      = model;
+  else
+    babl->format.model      = (void*)babl_remodel_with_space ((void*)model, space);
+
   memcpy (babl->format.component, component, sizeof (BablComponent *) * components);
   memcpy (babl->format.type, type, sizeof (BablType *) * components);
   memcpy (babl->format.sampling, sampling, sizeof (BablSampling *) * components);
@@ -115,10 +125,38 @@ format_new (const char     *name,
   babl->format.visited = 0;
   babl->format.image_template = NULL;
   babl->format.format_n = 0;
+  babl->format.palette = 0;
 
+  babl->format.space = (void*)space;
+  babl->format.encoding = NULL;
+  babl->instance.doc = doc;
   return babl;
 }
 
+Babl *
+format_new_from_format_with_space (const Babl *format, 
+                                   const Babl *space)
+{
+  Babl *ret;
+  char new_name[256];
+  snprintf (new_name, sizeof (new_name), "%s-%s", babl_get_name ((void*)format),
+                                                  babl_get_name ((void*)space));
+  ret = babl_db_find (babl_format_db(), new_name);
+  if (ret)
+    return ret;
+
+  ret = format_new (new_name,
+                    0,
+                    format->format.planar, format->format.components,
+                    (void*)babl_remodel_with_space (BABL(format->format.model), space),
+                    space,
+                    format->format.component, format->format.sampling, (void*)format->format.type, NULL);
+
+  ret->format.encoding = babl_get_name(format);
+  babl_db_insert (db, (void*)ret);
+  return ret;
+}
+
 
 static char *
 create_name (const BablModel *model,
@@ -128,6 +166,7 @@ create_name (const BablModel *model,
 {
   char            buf[512] = "";
   char           *p = &buf[0];
+  ssize_t         left;
   int             i;
   int             same_types = 1;
   const BablType**t          = type;
@@ -135,9 +174,11 @@ create_name (const BablModel *model,
   BablComponent **c1         = component;
   BablComponent **c2         = model->component;
 
-
-  sprintf (p, "%s ", model->instance.name);
+  left = 512;
+  snprintf (p, left, "%s ", model->instance.name);
   p += strlen (model->instance.name) + 1;
+  left -= strlen (model->instance.name) + 1;
+  babl_assert (left >= 0);
 
   i = components;
   while (i--)
@@ -169,7 +210,7 @@ create_name (const BablModel *model,
 
   if (same_types)
     {
-      sprintf (p, "%s", first_type->instance.name);
+      snprintf (p, left, "%s", first_type->instance.name);
       return babl_strdup (buf);
     }
 
@@ -177,24 +218,26 @@ create_name (const BablModel *model,
 
   while (i--)
     {
-      sprintf (p, "(%s as %s) ",
+      snprintf (p, left, "(%s as %s) ",
                (*component)->instance.name,
                (*type)->instance.name);
       p += strlen ((*component)->instance.name) +
            strlen ((*type)->instance.name) + strlen ("( as ) ");
+      left -= strlen ((*component)->instance.name) +
+              strlen ((*type)->instance.name) + strlen ("( as ) ");
+      babl_assert (left >= 0);
       component++;
       type++;
     }
   return babl_strdup (buf);
 }
 
-
 static char *
 ncomponents_create_name (const Babl *type,
                          int         components)
 {
   char buf[512];
-  sprintf (buf, "%s[%i] ", type->instance.name, components);
+  snprintf (buf, sizeof (buf), "%s[%i] ", type->instance.name, components);
   return babl_strdup (buf);
 }
 
@@ -242,7 +285,8 @@ babl_format_n (const Babl *btype,
   babl = format_new (name,
                      id,
                      planar, components, model,
-                     component, sampling, type);
+                     babl_space("sRGB"),
+                     component, sampling, type, NULL);
 
   babl_format_set_is_format_n (babl);
 
@@ -262,12 +306,12 @@ babl_format_is_format_n (const Babl *format)
   return 0;
 }
 
-
 static int
 is_format_duplicate (Babl            *babl,
                      int              planar,
                      int              components,
                      BablModel       *model,
+                     const Babl      *space,
                      BablComponent  **component,
                      BablSampling   **sampling,
                      const BablType **type)
@@ -276,7 +320,8 @@ is_format_duplicate (Babl            *babl,
 
   if (babl->format.planar     != planar     ||
       babl->format.components != components ||
-      babl->format.model      != model)
+      babl->format.model      != model      ||
+      babl->format.space      != (void*)space)
     return 0;
 
   for (i = 0; i < components; i++)
@@ -299,6 +344,8 @@ babl_format_new (const void *first_arg,
   int            planar     = 0;
   int            components = 0;
   BablModel     *model      = NULL;
+  const Babl    *space      = babl_space ("sRGB");
+  const char    *doc        = NULL;
   BablComponent *component [BABL_MAX_COMPONENTS];
   BablSampling  *sampling  [BABL_MAX_COMPONENTS];
   const BablType*type      [BABL_MAX_COMPONENTS];
@@ -312,7 +359,34 @@ babl_format_new (const void *first_arg,
 
   while (1)
     {
-      if (BABL_IS_BABL (arg))
+      /* first, we assume arguments to be strings */
+      if (!strcmp (arg, "id"))
+        {
+          id = va_arg (varg, int);
+        }
+
+      else if (!strcmp (arg, "name"))
+        {
+          name = babl_strdup (va_arg (varg, char *));
+        }
+
+      else if (!strcmp (arg, "doc"))
+        {
+          doc = babl_strdup (va_arg (varg, const char *));
+        }
+
+      else if (!strcmp (arg, "packed"))
+        {
+          planar = 0;
+        }
+
+      else if (!strcmp (arg, "planar"))
+        {
+          planar = 1;
+        }
+
+      /* if we didn't point to a known string, we assume argument to be babl */
+      else if (BABL_IS_BABL (arg))
         {
           Babl *babl = (Babl *) arg;
 
@@ -346,6 +420,10 @@ babl_format_new (const void *first_arg,
                 current_sampling = (BablSampling *) arg;
                 break;
 
+              case BABL_SPACE:
+                space = (Babl*) arg;
+                break;
+
               case BABL_MODEL:
                 if (model)
                   {
@@ -375,26 +453,6 @@ babl_format_new (const void *first_arg,
                 break;
             }
         }
-      /* if we didn't point to a babl, we assume arguments to be strings */
-      else if (!strcmp (arg, "id"))
-        {
-          id = va_arg (varg, int);
-        }
-
-      else if (!strcmp (arg, "name"))
-        {
-          name = babl_strdup (va_arg (varg, char *));
-        }
-
-      else if (!strcmp (arg, "packed"))
-        {
-          planar = 0;
-        }
-
-      else if (!strcmp (arg, "planar"))
-        {
-          planar = 1;
-        }
 
       else
         {
@@ -411,6 +469,15 @@ babl_format_new (const void *first_arg,
   if (!name)
     name = create_name (model, components, component, type);
 
+  if (space != babl_space ("sRGB"))
+  {
+    char *new_name = babl_malloc (strlen (name) +
+                                  strlen (babl_get_name ((Babl*)space)) + 1);
+    sprintf (new_name, "%s-%s", name, babl_get_name ((Babl*)space));
+    babl_free (name);
+    name = new_name;
+  }
+
   if (!model)
     {
       babl_log ("no model specified for format '%s'", name);
@@ -434,7 +501,7 @@ babl_format_new (const void *first_arg,
       /* There is an instance already registered by the required id/name,
        * returning the preexistent one instead if it doesn't differ.
        */
-      if (!is_format_duplicate (babl, planar, components, model,
+      if(0)if (!is_format_duplicate (babl, planar, components, model, space,
                                 component, sampling, type))
         babl_fatal ("BablFormat '%s' already registered "
                     "with different content!", name);
@@ -443,10 +510,10 @@ babl_format_new (const void *first_arg,
       return babl;
     }
 
-  babl = format_new (name,
+  babl = format_new ((void*)name,
                      id,
-                     planar, components, model,
-                     component, sampling, type);
+                     planar, components, model, space,
+                     component, sampling, type, doc);
 
   babl_db_insert (db, babl);
   babl_free (name);
@@ -542,29 +609,10 @@ babl_format_with_model_as_type (const Babl *model,
   );
 }
 
-#define test_pixels    256
-
-static double *
-test_create (void)
-{
-  double *test;
-  int     i;
-
-  srandom (20050728);
-
-  test = babl_malloc (sizeof (double) * test_pixels * 4);
-
-  for (i = 0; i < test_pixels * 4; i++)
-    test [i] = (double) random () / RAND_MAX;
-
-  return test;
-}
-
 double
 babl_format_loss (const Babl *babl)
 {
   double  loss = 0.0;
-  double *test;
   void   *original;
   double *clipped;
   void   *destination;
@@ -575,6 +623,9 @@ babl_format_loss (const Babl *babl)
   Babl       *fish_to;
   Babl       *fish_from;
 
+  const double *test = babl_get_format_test_pixels ();
+  const int     test_pixels = babl_get_num_format_test_pixels ();
+
   ref_fmt = babl_format_new (
     babl_model ("RGBA"),
     babl_type ("double"),
@@ -591,7 +642,6 @@ babl_format_loss (const Babl *babl)
   fish_to   = babl_fish_reference (ref_fmt, fmt);
   fish_from = babl_fish_reference (fmt, ref_fmt);
 
-  test        = test_create ();
   original    = babl_calloc (test_pixels, fmt->format.bytes_per_pixel);
   clipped     = babl_calloc (test_pixels, ref_fmt->format.bytes_per_pixel);
   destination = babl_calloc (test_pixels, fmt->format.bytes_per_pixel);
@@ -604,8 +654,6 @@ babl_format_loss (const Babl *babl)
 
   loss = babl_rel_avg_error (clipped, test, test_pixels * 4);
 
-  fish_to->fish.processings   -= 2;
-  fish_from->fish.processings -= 2;
   fish_to->fish.pixels        -= test_pixels * 2;
   fish_from->fish.pixels      -= test_pixels * 2;
 
@@ -613,7 +661,6 @@ babl_format_loss (const Babl *babl)
   babl_free (clipped);
   babl_free (destination);
   babl_free (transformed);
-  babl_free (test);
 
   ((Babl*)babl)->format.loss = loss;
   return loss;
@@ -637,7 +684,8 @@ babl_get_user_data (const Babl *babl)
 }
 
 void
-babl_set_user_data (const Babl *cbabl, void *data)
+babl_set_user_data (const Babl *cbabl, 
+                    void       *data)
 {
   Babl *babl = (Babl*) cbabl;
   switch (cbabl->instance.class_type)
@@ -653,4 +701,88 @@ babl_set_user_data (const Babl *cbabl, void *data)
     }
 }
 
+const Babl *
+babl_format_get_model (const Babl *format)
+{
+  if (format->class_type == BABL_FORMAT)
+    {
+      return (Babl*)format->format.model;
+    }
+  return NULL;
+}
+
+const Babl * babl_format_get_space      (const Babl *format)
+{
+  if (format->class_type == BABL_FORMAT)
+    {
+      return (Babl*)format->format.space;
+    }
+  return NULL;
+}
+
 BABL_CLASS_IMPLEMENT (format)
+
+const char *
+babl_format_get_encoding (const Babl *babl)
+{
+  if (!babl) return NULL;
+  if (babl->format.encoding) return babl->format.encoding;
+  return babl_get_name (babl);
+}
+
+const Babl *
+babl_format_with_space (const char *encoding, const Babl *space)
+{
+  const Babl *example_format = (void*) encoding;
+  if (!encoding) return NULL;
+
+  if (BABL_IS_BABL (example_format))
+  {
+    encoding = babl_get_name (example_format);
+    if (babl_format_get_space (example_format) != babl_space ("sRGB"))
+    {
+      encoding = babl_format_get_encoding (example_format);
+    }
+  }
+
+  if (!space)
+    space = babl_space ("sRGB");
+
+  if (space->class_type == BABL_FORMAT)
+  {
+    space = space->format.space;
+  }
+  else if (space->class_type == BABL_MODEL)
+  {
+    space = space->model.space;
+  }
+  else if (space->class_type != BABL_SPACE)
+  {
+    return NULL;
+  }
+  example_format = babl_format (encoding);
+
+  if (space == babl_space("sRGB"))
+    return example_format;
+
+  if (babl_format_is_palette (example_format))
+  {
+    /* XXX we should allocate a new palette name, and 
+           duplicate the path data, converted for new space
+     */
+    return example_format;
+  }
+
+  return format_new_from_format_with_space (example_format, space);
+}
+
+int
+babl_format_exists (const char *name)
+{
+  if (babl_db_exist_by_name (db, name))
+    return 1;
+  return 0;
+}
+
+
+
